/*
 * (c) copyright 2017-2020 Universidade de Vigo. All rights reserved.
 * formella@uvigo.es, http://formella.webs.uvigo.es
 */

/**
  \file
  \brief Multiplication as an example
*/

class Mul implements Runnable {
  private final int id;
  private final int n;
  private Int p;
  private Int q;
  private Int r;

  Mul(int id, int n, Int p, Int q, Int r) {
    this.id=id;
    this.n=n;
    this.p=p;
    this.q=q;
    this.r=r;
  }

  public void run() {
    try {
      final Int minusOne=new Int(-1);
        // Here, we check "greater than n" to avoid negative q !!
      while(q.Get()>n) {
        r.Add(p);
        q.Add(minusOne);
      }
    }
    catch(Exception E) {
      System.out.println("some error..."+id);
    }
    // We don't write final message, so we don't disturb measurement.
    // finally {
      // System.out.println("finished ..."+id);
    // }
  }
}

/// \brief Main class to launch application.
/// \note We always print a final message before leaving.
class Multi {
  public static void main(
    String[] args
  ) {
    try {
      if(args.length<3) {
        System.out.println("# provide at least 3 arguments: p q n");
        System.exit(1);
      }

      System.out.println(
        "# starting... "+args[0]+"*"+args[1]+" with "+args[2]+" threads"
      );
      final int n=Integer.parseInt(args[2]);

      double t=0.0;
      for(int j=0; j<100; ++j) {
        Timer T=new Timer(1);
        T.Start(0);
        Thread[] threads=new Thread[n];

        Int p=new Int(Integer.parseInt(args[0]));
        Int q=new Int(Integer.parseInt(args[1]));
        Int r=new Int(0);

        for(int i=0; i<threads.length; ++i) {
          threads[i]=new Thread(new Mul(i,n,p,q,r));
        }
        for(int i=0; i<threads.length; ++i) {
          threads[i].start();
        }
        for(int i=0; i<threads.length; ++i) {
          threads[i].join();
        }
          // Here we take care of the left-overs.
        final Int minusOne=new Int(-1);
        while(q.Get()>0) {
          r.Add(p);
          q.Add(minusOne);
        }

        T.Stop(0);
        t+=T.Elapsed();
          // Check whether we multiplied correctly...
        if(Integer.parseInt(args[0])*Integer.parseInt(args[1])!=r.Get()) {
          System.out.println("oh,oh, error calculating...");
        }
      }
      System.out.println(n+", "+t/10);
    }
    catch(Exception E) {
      System.out.println("# some exception...");
      System.out.println("# "+E.getClass().getCanonicalName());
      System.exit(1);
    }
    finally {
      System.out.println("# exiting...");
    }
  }
}

