/*
 * (c) copyright 2017-2020 Universidade de Vigo. All rights reserved.
 * formella@uvigo.es, http://formella.webs.uvigo.es
 */

/**
  \file
  \brief Multiplication as an example
*/

class Mul implements Runnable {
  private final int id;
  private final int n;
  private int p;
  private int q;
  private Int r;

  Mul(int id, int n, Int p, Int q, Int r) {
    this.id=id;
    this.n=n;
    this.p=p.Get();
    this.q=q.Get()/n;
    this.r=r;
  }

  public void run() {
    try {
        // Here, we run on local variables.
      int local_r=0;
      while(q>0) {
        local_r+=p;
        --q;
      }
      final Int My_r=new Int(local_r);
      r.Add(My_r);
    }
    catch(Exception E) {
      System.out.println("some error..."+id);
    }
  }
}

/// \brief Main class to launch application.
/// \note We always print a final message before leaving.
class MultiP {
  public static void main(
    String[] args
  ) {
    try {
      if(args.length<3) {
        System.out.println("# provide 3 arguments...");
        System.exit(1);
      }

      System.out.println(
        "# starting... "+args[0]+"*"+args[1]+" with "+args[2]+" threads"
      );
      final int n=Integer.parseInt(args[2]);

      double t=0.0;
      final int loops=10;
      for(int j=0; j<loops; ++j) {
        Timer T=new Timer(1);
        T.Start(0);

        Int p=new Int(Integer.parseInt(args[0]));
        Int q=new Int(Integer.parseInt(args[1]));
        Int r=new Int(0);

        Thread[] threads=new Thread[n];

        for(int i=0; i<threads.length; ++i) {
          threads[i]=new Thread(new Mul(i,n,p,q,r));
        }
        for(int i=0; i<threads.length; ++i) {
          threads[i].start();
        }
          // Here we help with the left-overs.
        int local_p=p.Get();
        int local_q=q.Get()%n;
        int local_r=0;
        while(local_q>0) {
          local_r+=local_p;
          --local_q;
        }

        for(int i=0; i<threads.length; ++i) {
          threads[i].join();
        }
        final Int My_r=new Int(local_r);
        r.Add(My_r);

        T.Stop(0);
        t+=T.Elapsed();
        System.out.println(
          "# "+args[0]+"*"+args[1]+" (with "+args[2]+" threads) = "+r.Get()
        );
      }
      System.out.println(n+", "+t/loops);
    }
    catch(Exception E) {
      System.out.println("# some exception...");
      System.out.println("# "+E.getClass().getCanonicalName());
      System.exit(1);
    }
    finally {
      System.out.println("# exiting...");
    }
  }
}

